/* $Id: mmrecover.c 383 2009-08-16 09:21:26Z romain $ */

#include <err.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/endian.h>

#include "progress.h"

#define MARKER_SIZE 2
#define PAYLOAD_SIZE_SIZE 2

#define FILE_MAX_SIZE	1024*1024*10

#define JM_START	0xff
#define JM_SOI		0xd8
#define JM_SOF0		0xc0
#define JM_SOF2		0xc2
#define JM_DHT		0xc4
#define JM_DQT		0xdb
#define JM_DRI		0xdd
#define JM_SOS		0xda
#define JM_RSTn		0xd0
#define JM_APPn		0xe0
#define JM_COM		0xfe
#define JM_EOI		0xd9

#define LOG(log, status, start, offset, saved_file_name, fmt, ...) do { \
    fprintf(log, "%d:%s:0x%016llx:0x%016llx:0x%08llx:%s:" fmt, \
	status, FileName, start, offset, offset - start, saved_file_name, \
	##__VA_ARGS__); \
    } while (0)

char *FileName;

int
read_jpeg(int stream, off_t offset, FILE *log)
{
    int size;

    uint8_t marker[MARKER_SIZE];
    uint8_t payload_size[PAYLOAD_SIZE_SIZE];

    uint8_t buffer[BUFSIZ];

    off_t image_start = offset;

    /* Read Markers */

    if (MARKER_SIZE != pread(stream, marker, sizeof marker, offset))
	goto error;
    offset += MARKER_SIZE;

    if (JM_START != marker[0] || JM_SOI != marker[1])
	return 0;

    while (JM_START == marker[0]) {
#if defined(DEBUG)
	fprintf(stdout, "Found JPEG Marker 0x%02x%02x at offset=0x%llx\n", marker[0], marker[1], offset);
#endif
	if ((JM_SOI == marker[1]) ||
	    (((marker[1] & 0xf0) == JM_RSTn) && ((marker[1] & 0x0f) <= 7)));
	    /* Nothing */

	else if ((marker[1] == JM_SOF0) ||
		(marker[1] == JM_SOF2) ||
		(marker[1] == JM_DHT) ||
		(marker[1] == JM_DQT) ||
		(marker[1] == JM_DRI) ||
		(marker[1] == JM_SOS) ||
		((marker[1] & 0xf0) == JM_APPn) ||
		(marker[1] == JM_COM)) {

	    if (PAYLOAD_SIZE_SIZE != pread(stream, payload_size, sizeof payload_size, offset))
		goto error;

	    /* Read marker payload size */
	    size = le16toh((payload_size[0] << 8) + payload_size[1]);
#if defined(DEBUG)
	    fprintf(stdout, "  Payload size: 0x%02x%02x (%d)\n", payload_size[0], payload_size[1], size);
#endif
	    offset += size;
	} else {
	    LOG(log, 1, image_start, offset, "-", "Found unrecognised JPEG Marker 0x%02x%02x", marker[0], marker[1]);
	    return -1;
	}

	if (marker[1] == JM_SOS)
	    break;

	if (MARKER_SIZE != pread(stream, marker, sizeof marker, offset))
	    goto error;
	offset += MARKER_SIZE;
    }

    if (marker[1] != JM_SOS) {
	LOG(log, 2, image_start, offset, "-", "JPEG Markers don't end with a Start Of Scan marker but 0x%02x%02x. Probably not an image.\n", marker[0], marker[1]);
	return -1;
    }

    /* Read entropy coded data */
    ssize_t br;
#if defined(DEBUG)
    fprintf(stderr, "Reading JPEG entropy coded data (buffer size=%d)\n", sizeof buffer);
#endif
    while ((0 < (br = pread(stream, buffer, sizeof buffer, offset))) && (offset - image_start < FILE_MAX_SIZE)) {
	ssize_t i;

	for (i=0;i<br-1;i++) {
	    if ((buffer[i] == JM_START) && (buffer[i+1] == JM_EOI)) {
#if defined(DEBUG)
		fprintf(stdout, "EOI Maker found at offset=0x%llx+0x%x\n", offset, i);
#endif
		return offset + i + MARKER_SIZE - image_start;
	    }
	}

    	offset += br;
    }

    LOG(log, 3, image_start, offset, "-", "No EOI Marker found after reading %lld+%d bytes of entropy encoded data.\n", offset - image_start, br);
    return -1;
error:
    perror("pread");
    fprintf(stderr, "Read error at offset=%llx (%lld) stream=%d\n", offset, offset, stream);
    exit(EXIT_FAILURE);
}

int
main(int argc, char *argv[])
{
    off_t offset = 0;

    int image_count = 0;
    char *log_filename = "mmrecover.log";
    char *filename_pattern = "recover_%06d.jpg";
    char filename_buf[BUFSIZ];

    /*
     * Parse command line arguments.
     */
    int ch;
    while ((ch = getopt(argc, argv, "l:o:p:")) != -1) {
	switch (ch) {
	    case 'l':
		log_filename = optarg;
		break;
	    case 'o':
		if (1 != sscanf(optarg, "%lld", &offset)) {
		    errx(EXIT_FAILURE, "Cannot parse \"%s\" as a number.", optarg);
		}
		break;
	    case 'p':
		filename_pattern = optarg;
		break;
	    default:
		exit(EXIT_FAILURE);
	}
    }

    argc -= optind;
    argv += optind;

    if (offset && (argc > 1)) {
    	errx(EXIT_FAILURE, "The -o option cannot be use with multiple input files.");
    }

    /* Open log */
    FILE *log = fopen(log_filename, "w");
    if (NULL == log) {
    	err(EXIT_FAILURE, "Failed to open %s for writing", log_filename);
    }

    fprintf(stdout, "Logging recovery session to %s.\n", log_filename);

    int fd;
    int n;
    struct stat sb;

    for (n=0; n<argc; n++) {

	FileName = argv[n];
	fd = open(FileName, O_RDONLY);
	if (fd == -1) {
	    err(EXIT_FAILURE, "Failed to open %s for reading", FileName);
	}

	int size = 0;
	if (-1 == fstat(fd, &sb)) {
	    err(EXIT_FAILURE, "Unable to stat %s", FileName);
	}
	fprintf(stdout, "Looking for JPEG images in %s...\n", FileName);
	lseek(fd, offset,  SEEK_SET);
	while (offset < sb.st_size) {
	    size = read_jpeg(fd, offset, log);
	    if (size > 0) {
		/* fprintf(stdout, "JPEG file of size %d found at offset=0x%llx\n", size, offset); */

		sprintf(filename_buf, filename_pattern, ++image_count);
		uint8_t *raw_data = malloc(size);
		if (size != pread(fd, raw_data, size, offset)) {
		    errx(EXIT_FAILURE, "Image beyond eof!\n");
		}

		int retry;
		do {
		    retry = 0;
		    int image_fd = open(filename_buf, O_WRONLY | O_CREAT | O_TRUNC, 0666);
		    if (image_fd < 0) {
			fprintf(stderr, "\n");
			warn("*** Failed to open %s for writing", filename_buf);
			retry = 1;
		    } else {
			if (size != write(image_fd, raw_data, size)) {
			    fprintf(stderr, "\n");
			    warn("*** Failed to write to %s", filename_buf);
			    retry = 1;
			}
			close(image_fd);
		    }

		    /* Wrte failed. Let the user investigate this an retry */
		    if (retry) {
			char answer[BUFSIZ];
		    	fprintf(stdout, "Do you want to retry (yes/no) ? ");
			fgets(answer, sizeof answer, stdin);
			retry = answer[0] != 'n';
		    }
		} while (retry);
		free(raw_data);

		LOG(log, 0, offset, offset + size, filename_buf, "JPEG image found saved as %s (%d bytes).\n", filename_buf, size);

		offset += (size / 512 + 1) * 512;
	    } else
		offset += 512;

	    progress_step((size) ? ((size == -1) ? 'x' : '.') : '\0');

	}
	progress_end();
	fprintf(stdout, "Found %d image(s)\n", image_count);
	close(fd);
	offset = 0;
    }

    fclose(log);

    exit(EXIT_SUCCESS);
} /* main() */

